""" Follow a QNM labeled by (s,l,m,n) as spin varies from a=0 upwards.

TODO Documentation.
"""


from __future__ import division, print_function, absolute_import

import logging

import numpy as np
from scipy import optimize, interpolate

from .angular import l_min, swsphericalh_A
from .nearby import NearbyRootFinder

from .schwarzschild.approx import Schw_QNM_estimate
from .schwarzschild.tabulated import QNMDict

# TODO some documentation here, better documentation throughout

class KerrSpinSeq(object):
    """Object to follow a QNM up a sequence in a, starting from
    a=0. Values for omega and the separation constant from one
    value of a are used to seed the root finding for the next
    value of a, to maintain continuity in a when separation
    constant order can change. Uses NearbyRootFinder to actually
    perform the root-finding.

    Parameters
    ----------
    a_max: float [default: .9]
      Maximum dimensionless spin of black hole for the sequence,
      0 <= a_max < 1.

    delta_a: float [default: 0.005]
      Step size in a for following the sequence from a=0 to a_max

    s: int [default: 2]
      Spin of field of interest

    m: int [default: 2]
      Azimuthal number of mode of interest

    l: int [default: 2]
      The l-number of a sequence starting from the
      analytically-known value at a=0

    l_max: int [default: 20]
      Maximum value of l to include in the spherical-spheroidal
      matrix for finding separation constant and mixing
      coefficients. Must be sufficiently larger than l of interest
      that angular spectral method can converge. The number of
      l's needed for convergence depends on a.

    omega_guess: complex [default: from schwarzschild.QNMDict]
      Initial guess of omega for root-finding

    tol: float [default: 1e-10]
      Tolerance for root-finding

    n: int [default: 0]
      Overtone number of interest (sets the inversion number for
      infinite continued fraction in Leaver's method)

    Nr: int [default: 300]
      Truncation number of radial infinite continued
      fraction. Must be sufficiently large for convergence.

    Nr_min: int [default: Nr]
      Minimum number of terms for evaluating continued fraction.

    Nr_max: int [default: 4000]
      Maximum number of terms for evaluating continued fraction.

    r_N: complex [default: 0.j]
      Seed value taken for truncation of infinite continued
      fraction. UNUSED, REMOVE

    """

    def __init__(self, *args, **kwargs):

        # Read args
        self.a_max       = kwargs.get('a_max',       0.9)
        # TODO Maybe change this to delta_a0
        self.delta_a     = kwargs.get('delta_a',     0.005)
        self.delta_a_min = 1.e-5 # TODO get rid of magic number
        self.delta_a_max = 4.e-3 # TODO get rid of magic number
        self.s           = kwargs.get('s',           -2)
        self.m           = kwargs.get('m',           2)
        self.l           = kwargs.get('l',           2)
        self.l_max       = kwargs.get('l_max',       20)
        self.tol         = kwargs.get('tol',         1e-10)
        self.n           = kwargs.get('n',           0)

        if ('omega_guess' in kwargs.keys()):
            self.omega_guess = kwargs.get('omega_guess')
        else:
            qnm_dict = QNMDict()
            self.omega_guess = qnm_dict(self.s, self.l, self.n)[0]

        self.Nr          = kwargs.get('Nr',          300)
        self.Nr_min      = self.Nr
        self.Nr_max      = kwargs.get('Nr_max',      4000)
        self.r_N         = kwargs.get('r_N',         0.j)

        # TODO check that values make sense!!!
        assert self.a_max < 1., ("a_max={} must be < 1.".format(self.a_max))
        assert self.l >= l_min(self.s, self.m), ("l={} must be "
                                                 ">= l_min={}".format(
                                                     self.l,
                                                     l_min(self.s, self.m)))

        # Create array of a's, omega's, and A's
        self.a     = []
        self.omega = []
        self.cf_err= []
        self.n_frac= []
        self.A     = []
        self.C     = []

        self.delta_a_prop = []

        self._interp_o_r = None
        self._interp_o_i = None
        self._interp_A_r = None
        self._interp_A_i = None

        # We need and instance of root finder
        self.solver = NearbyRootFinder(s=self.s, m=self.m,
                                       l_max=self.l_max,
                                       tol=self.tol,
                                       n_inv=self.n, Nr=self.Nr,
                                       Nr_min=self.Nr_min,
                                       Nr_max=self.Nr_max,
                                       r_N=self.r_N)

    # Change this to *extending* the sequence so this can be reused
    def do_find_sequence(self):
        """ TODO Document """

        logging.info("l={}, m={}, n={} starting".format(
            self.l, self.m, self.n))
        
        i  = 0  # TODO Allow to start at other values
        _a = 0. # TODO Allow to start at other values

        # Initializing the sequence, start with guesses
        A0 = swsphericalh_A(self.s, self.l, self.m)
        omega_guess = self.omega_guess

        _delta_a = self.delta_a

        while _a <= self.a_max:

            self.solver.set_params(a=_a, A_closest_to=A0,
                                   omega_guess=omega_guess)

            # Flag: is the continued fraction expansion converging?
            cf_conv = False

            while not cf_conv:

                result = self.solver.do_solve()

                if (result is None):
                    raise optimize.nonlin.NoConvergence('Failed to find '
                                                        'QNM in sequence '
                                                        'at a={}'.format(_a))

                # TODO This probably doesn't belong here
                # Ensure we start on the "positive frequency"
                # sequence.  This only works for i==0 (a=0.) because
                # there the separation constant is real.
                if ((i == 0) and (np.real(result) < 0)):
                    result = -np.conjugate(result)


                cf_err, n_frac = self.solver.get_cf_err()

                # TODO ACTUALLY DO SOMETHING WITH THESE NUMBERS
                cf_conv = True

                if cf_conv:
                    # Done with this value of a
                    self.a.append(_a)
                    self.omega.append(result)
                    self.A.append(self.solver.A)
                    self.C.append(self.solver.C)
                    self.cf_err.append(cf_err)
                    self.n_frac.append(n_frac)
                else:
                    # For the next attempt, try starting where we
                    # ended up
                    self.solver.set_params(omega_guess=result,
                                           A_closest_to=self.solver.A)
                    # Now try again, because cf_conv is still False

            # We always try to get the a_max value. If that's the
            # value we just did, break out of the loop by hand
            if (_a == self.a_max):
                break

            # For the next value of a, start with a guess based on
            # the previously-computed values. When we have two or more
            # values, we can do a quadratic fit. Otherwise just start
            # at the same value.
            if (i < 2):
                omega_guess = self.omega[-1]
                A0          = self.A[-1]

                _a = _a + _delta_a
            else:

                # Build interpolants and allow extrapolation
                # Sadly, UnivariateSpline does not work on complex data
                interp_o_r = interpolate.UnivariateSpline(
                    self.a[-3:], np.real(self.omega[-3:]),
                    s=0, # No smoothing!
                    k=2, ext=0)

                interp_o_i = interpolate.UnivariateSpline(
                    self.a[-3:], np.imag(self.omega[-3:]),
                    s=0, # No smoothing!
                    k=2, ext=0)

                interp_A_r = interpolate.UnivariateSpline(
                    self.a[-3:], np.real(self.A[-3:]),
                    s=0, # No smoothing!
                    k=2, ext=0)

                interp_A_i = interpolate.UnivariateSpline(
                    self.a[-3:], np.imag(self.A[-3:]),
                    s=0, # No smoothing!
                    k=2, ext=0)

                if (True or (i > np.Inf)): # Only do the curvature estimate after a while
                    # Their second derivatives
                    d2_o_r = interp_o_r.derivative(2)
                    d2_o_i = interp_o_i.derivative(2)
                    d2_A_r = interp_A_r.derivative(2)
                    d2_A_i = interp_A_i.derivative(2)

                    # Estimate the a-curvature of the omega and A functions:
                    d2_o = np.abs(d2_o_r(_a) + 1.j*d2_o_i(_a))
                    d2_A = np.abs(d2_A_r(_a) + 1.j*d2_A_i(_a))

                    # Get the larger of the two a-curvatures
                    d2 = np.max([d2_o, d2_A])

                    # This combination has units of a. The numerator
                    # is an empirical fudge factor
                    _delta_a = 0.05/np.sqrt(d2)

                self.delta_a_prop.append(_delta_a)

                # Make sure it's between our min and max allowed step size
                _delta_a = np.max([self.delta_a_min, _delta_a])
                _delta_a = np.min([self.delta_a_max, _delta_a])

                _a = _a + _delta_a

                # Make sure we get the end point
                if (_a > self.a_max):
                    _a = self.a_max

                omega_guess = interp_o_r(_a) + 1.j*interp_o_i(_a)
                A0          = interp_A_r(_a) + 1.j*interp_A_i(_a)

            i = i+1
            # Go to the next iteration of the loop

        logging.info("l={}, m={}, n={} completed with {} points".format(
            self.l, self.m, self.n, len(self.a)))

        self.build_interps()

    def build_interps(self):
        """ TODO document """

        # TODO do we want to allow extrapolation?

        k=3 # cubic

        # Sadly, UnivariateSpline does not work on complex data
        self._interp_o_r = interpolate.UnivariateSpline(
            self.a, np.real(self.omega),
            s=0, # No smoothing!
            k=k, ext=0)

        self._interp_o_i = interpolate.UnivariateSpline(
            self.a, np.imag(self.omega),
            s=0, # No smoothing!
            k=k, ext=0)

        self._interp_A_r = interpolate.UnivariateSpline(
            self.a, np.real(self.A),
            s=0, # No smoothing!
            k=k, ext=0)

        self._interp_A_i = interpolate.UnivariateSpline(
            self.a, np.imag(self.A),
            s=0, # No smoothing!
            k=k, ext=0)

    def __call__(self, a):
        """ TODO document """

        # TODO validate input, 0 <= a < 1.
        # TODO if a > a_max then extend
        # TODO take parameter of whether to solve at guess or not

        o_r = self._interp_o_r(a)
        o_i = self._interp_o_i(a)

        A_r = self._interp_A_r(a)
        A_i = self._interp_A_i(a)

        omega_guess = complex(o_r, o_i)
        A_guess     = complex(A_r, A_i)

        self.solver.set_params(a=a, omega_guess=omega_guess,
                               A_closest_to=A_guess)

        result = self.solver.do_solve()

        if (result is None):
            raise optimize.nonlin.NoConvergence('Failed to find '
                                                'QNM in sequence '
                                                'at a={}'.format(a))

        cf_err, n_frac = self.solver.get_cf_err()

        # Do we want to insert these numbers into the arrays?

        return result, self.solver.A, self.solver.C



""" Computing, loading, and storing tabulated Schwarzschild QNMs.

TODO Documentation.
"""

from __future__ import division, print_function, absolute_import

import logging
import pickle
import os

import numpy as np

from ..angular import l_min
from .overtonesequence import SchwOvertoneSeq

# TODO some documentation here, better documentation throughout

def build_Schw_dict(*args, **kwargs):
    """ Function to build a dict of Schwarzschild QNMs.

    Loops over values of (s,l), using SchwOvertoneSeq to find
    sequences in n.

    TODO Documentation

    Parameters
    ----------
    s_arr: [int] [default: [-2, -1, 0]]
      Array of s values to run over.

    n_max: int [default: 20]
      Maximum overtone number to run over (inclusive).

    l_max: int [default: 20]
      Maximum angular harmonic number to run over (inclusive).

    tol: float [default: 1e-10]
      Tolerance to pass to SchwOvertoneSeq.

    Returns
    -------
    dict
      A dict with tuple keys (s,l,n).
      The value at d[s,l,n] is a tuple (omega, cf_err, n_frac)
      where omega is the frequency omega_{s,l,n}, cf_err is the
      estimated truncation error for the continued fraction, and
      n_frac is the depth of the continued fraction evaluation.

    """

    # TODO: Should we allow any other params to be customizable?
    s_arr   = kwargs.get('s_arr',   [-2, -1, 0])
    n_max   = kwargs.get('n_max',   20)
    l_max   = kwargs.get('l_max',   20)

    tol     = kwargs.get('tol',     1e-10)

    Schw_dict = {}
    Schw_err_dict = {}

    for s in s_arr:
        ls = np.arange(l_min(s,0),l_max+1)
        for l in ls:
            Schw_seq = SchwOvertoneSeq(s=s, l=l,
                                       n_max=n_max, tol=tol)
            try:
                Schw_seq.find_sequence()
            except:
                logging.warn("Failed at s={}, l={}".format(s, l))
            for n, (omega, cf_err, n_frac) in enumerate(zip(Schw_seq.omega,
                                                            Schw_seq.cf_err,
                                                            Schw_seq.n_frac)):
                Schw_dict[(s,l,n)] = (np.asscalar(omega),
                                      np.asscalar(cf_err),
                                      int(n_frac))
                Schw_dict[(-s,l,n)] = Schw_dict[(s,l,n)]

    return Schw_dict

############################################################

class QNMDict(object):
    """ Object for getting/holding/(pre-)computing Schwarzschild QNMs.

    This class uses the "borg" pattern, so the table of precomputed
    values will be shared amongst all instances of the class.  A set
    of precomputed QNMs can be loaded/stored from a pickle file with
    :meth:`load_dict()` and :meth:`write_dict()`.  The main interface
    is via the special :meth:`__call__()` method which is invoked via
    `object(s,l,n)`.  If the QNM labeled by (s,l,n) has already been
    computed, it will be returned.  Otherwise we try to compute it and
    then return it.

    Attributes
    ----------
    seq_dict: dict
      Keys are tuples (s,l) which label an overtone sequence of QNMs.
      Values are instances of
      :class:`qnm.schwarzschild.overtonesequence.SchwOvertoneSeq`.

    Parameters
    ----------
    init: bool, optional [default: False]
      Whether or not to call :meth:`load_dict()` when initializing
      this instance.

    dict_pickle_file: string, optional [default: None]
      Path to pickle file that holds precomputed QNMs. If the value is
      None, get the default from :meth:`default_pickle_file()`.

    """

    # Borg pattern, the QNM table will be shared among all instances
    _shared_state = {}

    def __init__(self, init=False, dict_pickle_file=None):

        self.__dict__ = self._shared_state

        if (not hasattr(self, 'seq_dict')):
            # First!
            self.seq_dict = {}

        if (init):
            self.load_dict(dict_pickle_file=dict_pickle_file)


    def default_pickle_file(self):
        """Give the default path of the QNM dict pickle file, `<dirname of this file>/data/Schw_dict.pickle`.

        Returns
        -------
        string
          `<dirname of this file>/data/Schw_dict.pickle`

        """

        return os.path.abspath(
            '{}/data/Schw_dict.pickle'.format(
                os.path.dirname(os.path.realpath(__file__))))

    def load_dict(self, dict_pickle_file=None):
        """ Load a Schw QNM dict from disk.

        Parameters
        ----------
        dict_pickle_file: string [default: from default_pickle_file]
          Filename for reading (or writing) dict of Schwarzschild QNMs

        """

        if (dict_pickle_file is None):
            dict_pickle_file = self.default_pickle_file()

        try:
            with open(dict_pickle_file, 'rb') as handle:
                logging.info("Loading Schw QNM dict from file {}".format(dict_pickle_file))
                self.seq_dict = pickle.load(handle)
        except UnicodeDecodeError as e:
            with open(dict_pickle_file, 'rb') as handle:
                logging.info("Loading Schw QNM dict from file {}".format(dict_pickle_file))
                self.seq_dict = pickle.load(handle, encoding='latin1')
        except:
            logging.info("Could not load Schw QNM dict from file {}".format(dict_pickle_file))

        return

    def write_dict(self, dict_pickle_file=None):
        """Write the current state of the QNM dict to disk.

        Parameters
        ----------
        dict_pickle_file: string [default: from default_pickle_file()]
          Filename for reading (or writing) dict of Schwarzschild QNMs
        """

        if (dict_pickle_file is None):
            dict_pickle_file = self.default_pickle_file()

        _the_dir = os.path.dirname(dict_pickle_file)
        if not os.path.exists(_the_dir):
            try:
                os.mkdir(_the_dir)
            except:
                logging.warn("Could not create dir {} to store Schw QNM dict".format(_the_dir))

        try:
            with open(dict_pickle_file, 'wb') as handle:
                logging.info("Writing Schw QNM dict to file {}".format(dict_pickle_file))
                pickle.dump(self.seq_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)
        except:
            logging.warn("Could not write Schw QNM dict to file {}".format(dict_pickle_file))

        return

    def __call__(self, s, l, n):
        """Get the Schwarzschild QNM labeled by (s,l,n).

        If the QNM has already been computed, immediately return that
        value.  If the (s,l) sequence is in the dict, but n has not
        been computed, extend the sequence to n and return the QNM.
        If the (s,l) sequence is not already in the dict, add it to
        the dict out to overtone n.

        Parameters
        ----------
        s: int
          Spin weight of the field.

        l: int
          Multipole number of the QNM.

        n: int
          Overtone number of the QNM.

        Returns
        -------
        (complex, double, int)
          The complex value is the QNM frequency.  The double is the
          estimated truncation error for the continued fraction.  The
          int is the depth of the continued fraction evaluation.

        Raises
        ------
        TODO

        """

        # TODO Check values of (s,l)
        s = -np.abs(s)

        if ((s,l) not in self.seq_dict.keys()):
            self.seq_dict[(s,l)] = SchwOvertoneSeq(s=s, l=l, n_max=n)

        seq = self.seq_dict[(s,l)]

        # This is idempotent, no danger in "extending" if the data
        # have already been computed. Go to n+2 to try to guard
        # against skipping an overtone in the sequence.
        # This can raise an exception.
        seq.extend(n_max=n+2)

        return (seq.omega[n], seq.cf_err[n], int(seq.n_frac[n]))



""" Solve the radial Teukolsky equation via Leaver's method.

TODO Documentation.

.. Note that numba's decorators confuse autodoc. Therefore you must
   update docs/_autosummary/qnm.radial.rst if you add any functions
   that are decorated by numba.

"""

from __future__ import division, print_function, absolute_import

import numba
import numpy as np

from .contfrac import lentz

# TODO some documentation here, better documentation throughout

@numba.jit(nopython=True)
def sing_pt_char_exps(omega, a, s, m):
    r""" Compute the three characteristic exponents of the singular points
    of the radial Teukolsky equation.

    We want ingoing at the outer horizon and outgoing at infinity. The
    choice of one of two possible characteristic exponents at the
    inner horizon doesn't affect the minimal solution in Leaver's
    method, so we just pick one. Thus our choices are, in the
    nomenclature of [1]_, :math:`(\zeta_+, \xi_-, \eta_+)`.

    Parameters
    ----------
    omega: complex
      The complex frequency in the ansatz for the solution of the
      radial Teukolsky equation.

    a: double
      Spin parameter of the black hole, 0. <= a < 1 .

    s: int
      Spin weight of the field (i.e. -2 for gravitational).

    m: int
      Azimuthal number for the perturbation.

    Returns
    -------
    (complex, complex, complex)
      :math:`(\zeta_+, \xi_-, \eta_+)`

    References
    ----------
    .. [1] GB Cook, M Zalutskiy, "Gravitational perturbations of the
       Kerr geometry: High-accuracy study," Phys. Rev. D 90, 124021
       (2014), https://arxiv.org/abs/1410.7698 .

    """

    root = np.sqrt(1. - a*a)
    r_p, r_m = 1. + root, 1. - root

    sigma_p = (2.*omega*r_p - m*a)/(2.*root)
    sigma_m = (2.*omega*r_m - m*a)/(2.*root)

    zeta = +1.j * omega        # This is the choice \zeta_+
    xi   = - s - 1.j * sigma_p # This is the choice \xi_-
    eta  = -1.j * sigma_m      # This is the choice \eta_+

    return zeta, xi, eta

@numba.jit(nopython=True)
def D_coeffs(omega, a, s, m, A):
    """ The D_0 through D_4 coefficients that enter into the radial
    infinite continued fraction, Eqs. (31) of [1]_ .


    Parameters
    ----------
    omega: complex
      The complex frequency in the ansatz for the solution of the
      radial Teukolsky equation.

    a: double
      Spin parameter of the black hole, 0. <= a < 1 .

    s: int
      Spin weight of the field (i.e. -2 for gravitational).

    m: int
      Azimuthal number for the perturbation.

    A: complex
      Separation constant between angular and radial ODEs.

    Returns
    -------
    array[5] of complex
      D_0 through D_4 .

    References
    ----------
    .. [1] GB Cook, M Zalutskiy, "Gravitational perturbations of the
       Kerr geometry: High-accuracy study," Phys. Rev. D 90, 124021
       (2014), https://arxiv.org/abs/1410.7698 .

    """

    zeta, xi, eta = sing_pt_char_exps(omega, a, s, m)

    root  = np.sqrt(1. - a*a)

    p     = root * zeta
    alpha = 1. + s + xi + eta - 2.*zeta + s # Because we took the root \zeta_+
    gamma = 1. + s + 2.*eta
    delta = 1. + s + 2.*xi
    sigma = (A + a*a*omega*omega - 8.*omega*omega
             + p * (2.*alpha + gamma - delta)
             + (1. + s - 0.5*(gamma + delta))
             * (s + 0.5*(gamma + delta)))

    D = [0.j] * 5
    D[0] = delta
    D[1] = 4.*p - 2.*alpha + gamma - delta - 2.
    D[2] = 2.*alpha - gamma + 2.
    D[3] = alpha*(4.*p - delta) - sigma
    D[4] = alpha*(alpha - gamma + 1.)

    return D

def leaver_cf_trunc_inversion(omega, a, s, m, A,
                              n_inv, N=300, r_N=1.):
    """ Legacy function.

    Approximate the n_inv inversion of the infinite continued
    fraction for solving the radial Teukolsky equation, using
    N terms total for the approximation. This uses "bottom up"
    evaluation, and you can pass a seed value r_N to assume for
    the rest of the infinite fraction which has been truncated.
    The value returned is Eq. (44) of [1]_.

    Parameters
    ----------
    omega: complex
      The complex frequency for evaluating the infinite continued
      fraction.

    a: float
      Spin parameter of the black hole, 0. <= a < 1 .

    s: int
      Spin weight of the field (i.e. -2 for gravitational).

    m: int
      Azimuthal number for the perturbation.

    A: complex
      Separation constant between angular and radial ODEs.

    n_inv: int
      Inversion number for the infinite continued fraction. Finding
      the nth overtone is typically most stable when n_inv = n .

    N: int, optional [default: 300]
      The depth where the infinite continued fraction is truncated.

    r_N: float, optional [default: 1.]
      Value to assume for the rest of the infinite continued fraction
      past the point of truncation.

    Returns
    -------
    complex
      The nth inversion of the infinite continued fraction evaluated
      with these arguments.

    References
    ----------
    .. [1] GB Cook, M Zalutskiy, "Gravitational perturbations of the
       Kerr geometry: High-accuracy study," Phys. Rev. D 90, 124021
       (2014), https://arxiv.org/abs/1410.7698 .

    """

    n = np.arange(0, N+1)

    D = D_coeffs(omega, a, s, m, A)

    alpha =     n*n + (D[0] + 1.)*n + D[0]
    beta  = -2.*n*n + (D[1] + 2.)*n + D[3]
    gamma =     n*n + (D[2] - 3.)*n + D[4] - D[2] + 2.

    conv1 = 0.
    for i in range(0, n_inv): # n_inv is not included
        conv1 = alpha[i] / (beta[i] - gamma[i] * conv1)

    conv2 = -r_N # Is this sign correct?
    for i in range(N, n_inv, -1): # n_inv is not included
        conv2 = gamma[i] / (beta[i] - alpha[i] * conv2)

    return (beta[n_inv]
            - gamma[n_inv] * conv1
            - alpha[n_inv] * conv2)

# TODO possible choices for r_N: 0., 1., approximation using (34)-(38)

def leaver_cf_inv_lentz_old(omega, a, s, m, A, n_inv,
                            tol=1.e-10, N_min=0, N_max=np.Inf):
    """ Legacy function. Same as :meth:`leaver_cf_inv_lentz` except
    calling :meth:`qnm.contfrac.lentz` with temporary functions that
    are defined inside this function. Numba does not speed up
    this type of code. However it remains here for testing purposes.
    See documentation for :meth:`leaver_cf_inv_lentz` for parameters
    and return value.

    Examples
    --------

    >>> from qnm.radial import leaver_cf_inv_lentz_old, leaver_cf_inv_lentz
    >>> print(leaver_cf_inv_lentz_old(omega=.4 - 0.2j, a=0.02, s=-2, m=2, A=4.+0.j, n_inv=0))
    ((-3.5662773770495546-1.5388710793384461j), 9.702542314939062e-11, 76)

    Compare the two versions of the function:

    >>> old = leaver_cf_inv_lentz_old(omega=.4 - 0.2j, a=0.02, s=-2, m=2, A=4.+0.j, n_inv=0)
    >>> new = leaver_cf_inv_lentz(omega=.4 - 0.2j, a=0.02, s=-2, m=2, A=4.+0.j, n_inv=0)
    >>> [ old[i]-new[i] for i in range(3)]
    [0j, 0.0, 0]

    """

    D = D_coeffs(omega, a, s, m, A)

    # This is only use for the terminating fraction
    n = np.arange(0, n_inv+1)
    alpha =     n*n + (D[0] + 1.)*n + D[0]
    beta  = -2.*n*n + (D[1] + 2.)*n + D[3]
    gamma =     n*n + (D[2] - 3.)*n + D[4] - D[2] + 2.

    conv1 = 0.
    for i in range(0, n_inv): # n_inv is not included
        conv1 = alpha[i] / (beta[i] - gamma[i] * conv1)

    # In defining the below a, b sequences, I have cleared a fraction
    # compared to the usual way of writing the radial infinite
    # continued fraction. The point of doing this was that so both
    # terms, a(n) and b(n), tend to 1 as n goes to infinity. Further,
    # We can analytically divide through by n in the numerator and
    # denominator to make the numbers closer to 1.
    def a(i):
        n = i + n_inv - 1
        return -(n*n + (D[0] + 1.)*n + D[0])/(n*n + (D[2] - 3.)*n + D[4] - D[2] + 2.)

    def b(i):
        if (i==0): return 0
        n = i + n_inv
        return (-2.*n*n + (D[1] + 2.)*n + D[3])/(n*n + (D[2] - 3.)*n + D[4] - D[2] + 2.)

    conv2, cf_err, n_frac = lentz(a, b, tol=tol, N_min=N_min, N_max=N_max)

    return (beta[n_inv]
            - gamma[n_inv] * conv1
            + gamma[n_inv] * conv2), cf_err, n_frac


@numba.jit(nopython=True)
def leaver_cf_inv_lentz(omega, a, s, m, A, n_inv,
                               tol=1.e-10, N_min=0, N_max=np.Inf):
    """ Compute the n_inv inversion of the infinite continued
    fraction for solving the radial Teukolsky equation, using
    modified Lentz's method.
    The value returned is Eq. (44) of [1]_.

    Same as :meth:`leaver_cf_inv_lentz_old`, but with Lentz's method
    inlined so that numba can speed things up.

    Parameters
    ----------
    omega: complex
      The complex frequency for evaluating the infinite continued
      fraction.

    a: float
      Spin parameter of the black hole, 0. <= a < 1 .

    s: int
      Spin weight of the field (i.e. -2 for gravitational).

    m: int
      Azimuthal number for the perturbation.

    A: complex
      Separation constant between angular and radial ODEs.

    n_inv: int
      Inversion number for the infinite continued fraction. Finding
      the nth overtone is typically most stable when n_inv = n .

    tol: float, optional [default: 1.e-10]
      Tolerance for termination of Lentz's method.

    N_min: int, optional [default: 0]
      Minimum number of iterations through Lentz's method.

    N_max: int or comparable, optional [default: np.Inf]
      Maximum number of iterations for Lentz's method.

    Returns
    -------
    (complex, float, int)
      The first value (complex) is the nth inversion of the infinite
      continued fraction evaluated with these arguments. The second
      value (float) is the estimated error from Lentz's method. The
      third value (int) is the number of iterations of Lentz's method.

    Examples
    --------

    >>> from qnm.radial import leaver_cf_inv_lentz
    >>> print(leaver_cf_inv_lentz(omega=.4 - 0.2j, a=0.02, s=-2, m=2, A=4.+0.j, n_inv=0))
    ((-3.5662773770495546-1.5388710793384461j), 9.702542314939062e-11, 76)

    References
    ----------
    .. [1] GB Cook, M Zalutskiy, "Gravitational perturbations of the
       Kerr geometry: High-accuracy study," Phys. Rev. D 90, 124021
       (2014), https://arxiv.org/abs/1410.7698 .

    """

    D = D_coeffs(omega, a, s, m, A)

    # This is only use for the terminating fraction
    n = np.arange(0, n_inv+1)
    alpha =     n*n + (D[0] + 1.)*n + D[0]
    beta  = -2.*n*n + (D[1] + 2.)*n + D[3]
    gamma =     n*n + (D[2] - 3.)*n + D[4] - D[2] + 2.

    conv1 = 0.
    for i in range(0, n_inv): # n_inv is not included
        conv1 = alpha[i] / (beta[i] - gamma[i] * conv1)

    ##############################
    # Beginning of Lentz's method, inlined

    # TODO should tiny be a parameter?
    tiny = 1.e-30

    # This is starting with b_0 = 0 for the infinite continued
    # fraction. I could have started with other values (e.g. b_i
    # evaluated with i=0) but then I would have had to subtract that
    # same quantity away from the final result. I don't know if this
    # affects convergence.
    f_old = tiny

    C_old = f_old
    D_old = 0.

    conv = False

    j = 1
    n = n_inv

    while ((not conv) and (j < N_max)):

        # In defining the below a, b sequences, I have cleared a fraction
        # compared to the usual way of writing the radial infinite
        # continued fraction. The point of doing this was that so both
        # terms, a(n) and b(n), tend to 1 as n goes to infinity. Further,
        # We can analytically divide through by n in the numerator and
        # denominator to make the numbers closer to 1.
        an = -(n*n + (D[0] + 1.)*n + D[0])/(n*n + (D[2] - 3.)*n + D[4] - D[2] + 2.)
        n = n + 1
        bn = (-2.*n*n + (D[1] + 2.)*n + D[3])/(n*n + (D[2] - 3.)*n + D[4] - D[2] + 2.)

        D_new = bn + an * D_old

        if (D_new == 0):
            D_new = tiny

        C_new = bn + an / C_old

        if (C_new == 0):
            C_new = tiny

        D_new = 1./D_new
        Delta = C_new * D_new
        f_new = f_old * Delta

        if ((j > N_min) and (np.abs(Delta - 1.) < tol)): # converged
            conv = True

        # Set up for next iter
        j = j + 1
        D_old = D_new
        C_old = C_new
        f_old = f_new

    conv2 = f_new

    ##############################
    
    return (beta[n_inv]
            - gamma[n_inv] * conv1
            + gamma[n_inv] * conv2), np.abs(Delta-1.), j-1




""" Finding QNMs of Schwarzschild (numerically and approximately).

TODO Documentation.
"""

from __future__ import print_function, division, absolute_import

from . import approx
from . import tabulated
from . import overtonesequence



""" Solve the angular Teukolsky equation via spectral decomposition.

TODO Documentation.
"""

from __future__ import division, print_function, absolute_import

import numpy as np

# TODO some documentation here, better documentation throughout

def calF(s, l, m):
    """ Eq. (52b) """

    if ((0==s) and (0 == l+1)):
        # This can only happen when solving for the mode labeled by s=0, l=0, m=0
        return 0.

    return (np.sqrt( ((l+1)**2 - m*m) / (2*l+3) / (2*l+1) )
            * np.sqrt( ( (l+1)**2  - s*s)  / (l+1)**2 ))

def calG(s, l, m):
    """ Eq. (52c) """
    if (0 == l):
        return 0.

    return np.sqrt( ( l*l - m*m ) / (4*l*l - 1)) * np.sqrt(1 - s*s/l/l)

def calH(s, l, m):
    """ Eq. (52d) """
    if (0 == l) or (0 == s):
        return 0.

    return - m*s/l/(l+1)

def calA(s, l, m):
    """ Eq. (53a) """
    return calF(s,l,m) * calF(s,l+1,m)

def calD(s, l, m):
    """ Eq. (53b) """
    return calF(s,l,m) * (calH(s,l+1,m)  + calH(s,l,m))

def calB(s, l, m):
    """ Eq. (53c) """
    return (calF(s,l,m) * calG(s,l+1,m)
            + calG(s,l,m) * calF(s,l-1,m)
            + calH(s,l,m)**2)

def calE(s, l, m):
    """ Eq. (53d) """
    return calG(s,l,m) * (calH(s,l-1,m) + calH(s,l,m))

def calC(s, l, m):
    """ Eq. (53e) """
    return calG(s,l,m) * calG(s,l-1,m)

def swsphericalh_A(s, l, m):
    """ Angular separation constant at a=0.

    Eq. (50). Has no dependence on m. The formula is
      A_0 = l(l+1) - s(s+1)

    Parameters
    ----------
    s: integer
      Spin-weight of interest

    l: integer
      Angular quantum number of interest

    m: integer
      Magnetic quantum number, ignored

    Returns
    -------
    float
      Value of A(a=0) = l(l+1) - s(s+1)
    """

    return l*(l+1) - s*(s+1)

def M_matrix_elem(s, c, m, l, lprime):
    """ Eq. (55) """

    # Notice that the M matrix is pentadiagonal,
    # so no need to write the sub- and super-diagonals
    # separately
    if (lprime == l-2):
        return -c*c*calA(s,lprime,m)
    if (lprime == l-1):
        return (-c*c*calD(s,lprime,m)
                + 2*c*s*calF(s,lprime,m))
    if (lprime == l  ):
        return (swsphericalh_A(s,lprime,m)
                - c*c*calB(s,lprime,m)
                + 2*c*s*calH(s,lprime,m))
    if (lprime == l+1):
        return (-c*c*calE(s,lprime,m)
                + 2*c*s*calG(s,lprime,m))
    if (lprime == l+2):
        return -c*c*calC(s,lprime,m)

    return 0.

# I don't know if this is necessary ... can just iterate
def give_M_matrix_elem_ufunc(s, c, m):
    """ TODO Document """

    def elem(l, lprime):
        return M_matrix_elem(s, c, m, l, lprime)

    return np.frompyfunc(elem, 2, 1)

def l_min(s, m):
    """ TODO Documentation """

    return np.max([np.abs(s), np.abs(m)])

def M_matrix(s, c, m, l_max):
    """ TODO Document """

    ells = np.arange(l_min(s,m), l_max+1)

    uf = give_M_matrix_elem_ufunc(s, c, m)

    return uf.outer(ells,ells).astype(complex)

def sep_consts(s, c, m, l_max):
    """ TODO Document """

    return np.linalg.eigvals(M_matrix(s, c, m, l_max))

def sep_const_closest(A0, s, c, m, l_max):
    """ TODO Document """

    As = sep_consts(s, c, m, l_max)
    i_closest = np.argmin(np.abs(As-A0))
    return As[i_closest]

def C_and_sep_const_closest(A0, s, c, m, l_max):
    """ TODO Document """

    As, Cs = np.linalg.eig(M_matrix(s, c, m, l_max))
    i_closest = np.argmin(np.abs(As-A0))
    return As[i_closest], Cs[:,i_closest]



""" Analytic approximations for Schwarzschild QNMs.

The approximations implemented in this module can be used as initial
guesses when numerically searching for QNM frequencies.

"""

from __future__ import division, print_function, absolute_import

import numpy as np

def dolan_ottewill_expansion(s, l, n):
    """ High l asymptotic expansion of Schwarzschild QNM frequency.

    The result of [1]_ is an expansion in inverse powers of L =
    (l+1/2). Their paper stated this series out to L^{-4}, which is
    how many terms are implemented here. The coefficients in this
    series are themselves positive powers of N = (n+1/2). This means
    the expansion breaks down for large N.

    Parameters
    ----------
    s: int
      Spin weight of the field of interest.

    l: int
      Multipole number of interest.

    [The m parameter is omitted because this is just for Schwarzschild.]

    n: int
      Overtone number of interest.

    Returns
    -------
    complex
      Analytic approximation of QNM of interest.

    References
    ----------
    .. [1] SR Dolan, AC Ottewill, "On an expansion method for black
       hole quasinormal modes and Regge poles," CQG 26 225003 (2009),
       https://arxiv.org/abs/0908.0329 .
    """

    L    = l + 0.5
    N    = n + 0.5
    beta = 1. - s*s

    DO_coeffs = { -1: 1.,
                   0: -1.j * N,
                  +1: beta/3. - 5.*N*N/36. - 115./432.,
                  +2: -1.j * N * (beta/9. + 235.*N*N/3888. - 1415./15552.),
                  +3: (-beta*beta/27. + (204.*N*N + 211.)/3888.*beta
                       +(854160.*N**4 - 1664760.*N*N - 776939.)/40310784.),
                  +4: -1.j * N *(beta*beta/27. + (1100.*N*N-2719.)/46656.*beta
                                 + (11273136.*N**4 - 52753800.*N*N + 66480535.)/2902376448.) }

    omega = 0.
    for k, c in DO_coeffs.items():
        omega = omega + c * np.power(L, -k)

    omega = omega / np.sqrt(27)

    return omega

def large_overtone_expansion(s, l, n):
    r""" The eikonal approximation for QNMs, valid for l >> n >> 1 .

    This is just the first two terms of the series in
    :meth:`dolan_ottewill_expansion`.
    The earliest work I know deriving this result is [1]_ but there
    may be others. In the eikonal approximation, valid when
    :math:`l \gg n \gg 1`, the QNM frequency is

    .. math:: \sqrt{27} M \omega \approx (l+\frac{1}{2}) - i (n+\frac{1}{2}) .

    Parameters
    ----------
    s: int
      Spin weight of the field of interest.

    l: int
      Multipole number of interest.

    [The m parameter is omitted because this is just for Schwarzschild.]

    n: int
      Overtone number of interest.

    Returns
    -------
    complex
      Analytic approximation of QNM of interest.

    References
    ----------
    .. [1] V Ferrari, B Mashhoon, "New approach to the quasinormal
       modes of a black hole," Phys. Rev. D 30, 295 (1984)

    """

    k = np.log(3.)/(8. * np.pi)
    kappa = 0.25 # Surface gravity

    return k - 1.j * kappa * (n + 0.5)

def Schw_QNM_estimate(s, l, n):
    """ Give either :meth:`large_overtone_expansion` or :meth:`dolan_ottewill_expansion`.

    The Dolan-Ottewill expansion includes terms with higher powers of
    the overtone number n, so it breaks down faster at high n.

    Parameters
    ----------
    s: int
      Spin weight of the field of interest.

    l: int
      Multipole number of interest.

    [The m parameter is omitted because this is just for Schwarzschild.]

    n: int
      Overtone number of interest.

    Returns
    -------
    complex
      Analytic approximation of QNM of interest.

    """

    if (( n > 3 ) and (n >= 2*l)):
        return large_overtone_expansion(s, l, n)
    else:
        return dolan_ottewill_expansion(s, l, n)



""" Infinite continued fractions via Lentz's method.

TODO Documentation.
"""

from __future__ import division, print_function, absolute_import

import numpy as np

# TODO some documentation here, better documentation throughout

def lentz(a, b, tol=1.e-10, N_min=0, N_max=np.Inf, tiny=1.e-30):
    """ Compute a continued fraction via modified Lentz's method.

    This implementation is by the book [1]_.

    Parameters
    ----------
    a: callable returning numeric.
    b: callable returning numeric.

    tol: float [default: 1.e-10]
      Tolerance for termination of evaluation.

    N_min: int [default: 0]
      Minimum number of iterations to evaluate.

    N_max: int or comparable [default: np.Inf]
      Maximum number of iterations to evaluate.

    tiny: float [default: 1.e-30]
      Very small number to control convergence of Lentz's method when
      there is cancellation in a denominator.

    Returns
    -------
    (float, float, int)
      The first element of the tuple is the value of the continued
      fraction. The second element is the estimated error. The third
      element is the number of iterations.

    References
    ----------
    .. [1] WH Press, SA Teukolsky, WT Vetterling, BP Flannery,
       "Numerical Recipes," 3rd Ed., Cambridge University Press 2007,
       ISBN 0521880688, 9780521880688 .

    """

    f_old = b(0)

    if (f_old == 0):
        f_old = tiny

    C_old = f_old
    D_old = 0.

    conv = False

    j = 1

    while ((not conv) and (j < N_max)):

        aj, bj = a(j), b(j)

        D_new = bj + aj * D_old

        if (D_new == 0):
            D_new = tiny

        C_new = bj + aj / C_old

        if (C_new == 0):
            C_new = tiny

        D_new = 1./D_new
        Delta = C_new * D_new
        f_new = f_old * Delta

        if ((j > N_min) and (np.abs(Delta - 1.) < tol)): # converged
            conv = True

        # Set up for next iter
        j = j + 1
        D_old = D_new
        C_old = C_new
        f_old = f_new

    # Success or failure can be assessed by the user
    return f_new, np.abs(Delta - 1.), j-1

def lentz_gen(a, b, tol=1.e-10, N_min=0, N_max=np.Inf, tiny=1.e-30):
    """ Compute a continued fraction via modified Lentz's method,
    using generators rather than functions.

    This implementation is by the book [1]_.

    Parameters
    ----------
    a: generator yielding numeric.
    b: generator yielding numeric.

    tol: float [default: 1.e-10]
      Tolerance for termination of evaluation.

    N_min: int [default: 0]
      Minimum number of iterations to evaluate.

    N_max: int or comparable [default: np.Inf]
      Maximum number of iterations to evaluate.

    tiny: float [default: 1.e-30]
      Very small number to control convergence of Lentz's method when
      there is cancellation in a denominator.

    Returns
    -------
    (float, float, int)
      The first element of the tuple is the value of the continued
      fraction. The second element is the estimated error. The third
      element is the number of iterations.

    References
    ----------
    .. [1] WH Press, SA Teukolsky, WT Vetterling, BP Flannery,
       "Numerical Recipes," 3rd Ed., Cambridge University Press 2007,
       ISBN 0521880688, 9780521880688 .

    """

    f_old = next(b) # 0

    if (f_old == 0):
        f_old = tiny

    C_old = f_old
    D_old = 0.

    conv = False

    j = 1

    while ((not conv) and (j < N_max)):

        aj, bj = next(a), next(b) # j and j. 'a' started with 1.

        D_new = bj + aj * D_old

        if (D_new == 0):
            D_new = tiny

        C_new = bj + aj / C_old

        if (C_new == 0):
            C_new = tiny

        D_new = 1./D_new
        Delta = C_new * D_new
        f_new = f_old * Delta

        if ((j > N_min) and (np.abs(Delta - 1.) < tol)): # converged
            conv = True

        # Set up for next iter
        j = j + 1
        D_old = D_new
        C_old = C_new
        f_old = f_new

    # Success or failure can be assessed by the user
    return f_new, np.abs(Delta - 1.), j-1



'''
Standard setup.py to upload the code on pypi.

    python setup.py sdist bdist_wheel
    twine upload dist/*
'''
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

import sys
sys.path.append("qnm")

from _version import __version__

setuptools.setup(
    name="qnm",
    version=__version__,
    author="Leo C. Stein",
    author_email="leo.stein@gmail.com",
    description="Calculate quasinormal modes of Kerr black holes.",
    keywords='black holes quasinormal modes physics scientific computing numerical methods',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/duetosymmetry/qnm/",
    packages=setuptools.find_packages(),
    package_data={'qnm':['schwarzschild/data/*']}, # TODO
    include_package_data=True,
    install_requires=[
        'numpy',
        'scipy',
        'numba',
    ],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Astronomy",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Operating System :: OS Independent",
    ],
)
